"""Pyslabs numpy slab manipulation module


"""

import numpy as np
from io import BytesIO
from pyslabs.util import DEBUG_LEVEL, DEBUG_INFO


def length(slab, axis=0):
    return np.size(slab, axis)


def shape(ndarr):
    return ndarr.shape


def dump(path, ndarr):
    return np.save(path, ndarr)


def load(file):
    bio = BytesIO()
    bio.write(file.read())
    bio.seek(0) 
    return np.load(bio)


def get_slice(slab, key):
    return slab.__getitem__(tuple(key))


def stack(stacker, lower):

    if stacker is None:
        stacker = np.expand_dims(lower, axis=0)

    else:
        stacker = np.append(stacker, np.expand_dims(lower,0), 0)

    return stacker


def expand_dim(slab):
    return np.expand_dims(slab, axis=0)


def concatenate(concater, panel,  axis=0):

    if concater is None:
        return panel

    if type(concater) != type(panel):
        raise PE_Stabif_Typemismatch("%s != %s" % (type(concater).__name__,
                                    type(panel).__name__))

    return np.concatenate((concater, panel), axis=axis)


def squeeze(array):
    if DEBUG_LEVEL > DEBUG_INFO:
        print("Squeeze IN: ", array)

    is_squeezed = False

    if array.shape[0] == 1:
        array = array[0]
        is_squeezed = True

    if DEBUG_LEVEL > DEBUG_INFO:
        print("Squeeze OUT(array, is_squeezed): ", array, is_squeezed)

    return is_squeezed, array



"""Pyslabs error module

"""


class Pyslabs_Error(Exception):
    pass


class PE_Open_Unknownmode(Pyslabs_Error):
    pass


class PE_Init_Nobeginfile(Pyslabs_Error):
    pass


class PE_Slab_Shapemismatch(Pyslabs_Error):
    pass


class PE_Begin_Numproc(Pyslabs_Error):
    pass


class PE_Close_Numproc(Pyslabs_Error):
    pass


class PE_Close_Timeout(Pyslabs_Error):
    pass


class PE_Close_Shapemismatch(Pyslabs_Error):
    pass


class PE_Close_Unknowncheck(Pyslabs_Error):
    pass


class PE_Write_Duplicateslabfile(Pyslabs_Error):
    pass


class PE_Read_Exeedlength(Pyslabs_Error):
    pass


class PE_Stabif_Typemismatch(Pyslabs_Error):
    pass


class PE_Util_Typemismatch(Pyslabs_Error):
    pass


class PE_Close_Startindexerror(Pyslabs_Error):
    pass





"""Pyslabs slab read module


"""

from pyslabs import slabif
from pyslabs.error import PE_Read_Exeedlength


class VariableReaderV1():
    def __init__(self, tar_file, slab_tower, var_cfg, dim_cfg):

        self.tar_file = tar_file
        self.slab_tower = slab_tower
        self.dim_cfg = dim_cfg
        self.var_cfg = var_cfg
        self.array_shape = tuple(self.var_cfg["shape"])
        self.start = (0,) * len(self.array_shape)
        shape = []
        for s in self.array_shape:
            if s in self.dim_cfg:
                shape.append(self.dim_cfg[s]["length"])

            else:
                shape.append(s)
        self.shape = tuple(shape)

    def _get_slice(self, dim, st, so, se):

        st = self.start[dim] if st is None else st
        so = self.shape[dim] if so is None else so
        se = 1 if se is None else se
            
        if so in self.dim_cfg:
            so = self.dim_cfg[so]["length"]

        return slice(st, so, se)

    def __getitem__(self, key):

        whole = tuple([self._get_slice(dim, None, None, None)
                      for dim in range(len(self.shape))])

        is_slice = True
        nslices = 0

        if isinstance(key, int):
            key = (key,) + whole[1:]
            is_slice = False
            nslices = len(whole) - 1

        elif isinstance(key, slice):
            key = (self._get_slice(0, key.start, key.stop, key.step),) + whole[1:]
            nslices = len(whole)

        else:
            buf = []

            for i, k in enumerate(key):
                if isinstance(k, int):
                    buf.append(k)
                    if i==0:
                        is_slice = False

                elif isinstance(k, slice):
                    buf.append(self._get_slice(i, k.start, k.stop, k.step))
                    nslices += 1

            if len(self.shape) - len(key) > 0:
                nslices += (len(self.shape) - len(key))
                key = tuple(buf) + whole[len(key):len(self.shape)]

            else:
                key = tuple(buf)
#
#        shape = []
#        for s in self.shape[1:]:
#            if s in self.dim_cfg:
#                shape.append(self.dim_cfg[s]["length"])
#
#            else:
#                shape.append(s)

        is_squeezed, array = slabif.get_array(self.tar_file, self.slab_tower, self.shape[1:],
                                        key[1:], key[0])

        ndim = slabif.ndim(array)
        if ndim > nslices:
            _, array = slabif.squeeze(array)

        elif ndim < nslices:
            array = slabif.expand_dim(array)

        return array
#
#        if self.unstackable:
#            if not is_slice and len(array) == 1 and slabif.ndim(array) == len(self.shape):
#                print("BBBB", self.shape, key, array.shape, is_squeezed)
#                return array[0]
#            else:
#                return array
#
#        elif is_squeezed:
#            if is_slice:
#                print("AAAA", self.shape, key, slabif.shape(array), is_squeezed)
#                return slabif.exapand_dim(array)
#            else:
#                return array
#        else:
#            return array
#        if (self.unstackable and not is_slice and len(array) == 1 and
#            slabif.ndim(array) == len(self.shape)):
#            print("BBBB", self.shape, key, array.shape, is_squeezed)
#            return array[0]
#
#        else:
#            print("AAAA", self.shape, key, slabif.shape(array), is_squeezed)
#            return array
#



"""Pyslabs builtin-data slab manipulation module


"""

import os, io, pickle, itertools

from pyslabs.error import PE_Stabif_Typemismatch
from pyslabs.util import ScalarList, DEBUG_LEVEL, DEBUG_INFO

def length(slab, axis=0):

    return len(slab)


def slice2type(_slice, _type):

    return _type(_slice)


def shape(slab):

    s = []

    while slab:
        try:
            l = len(slab)

            if l > 0:
                s.append(l)
                slab = slab[0]

            else:
                break
        except TypeError:
            break

    return tuple(s)


def dump(path, slab):

    with io.open(path, "wb") as fp:
        pickle.dump(slab, fp)
        fp.flush()
        os.fsync(fp.fileno())

def load(tar_file):

    return pickle.load(tar_file)


def stack(stacker, lower):

    if stacker is None:
        try:
            return type(lower)([lower])

        except TypeError:
            return (lower,)

    if isinstance(stacker, list):
        stacker.append(lower)

    elif isinstance(stacker, tuple):
        stacker += (lower,)

    else:
        import pdb; pdb.set_trace()        

    return stacker


def expand_dim(array):

    return type(array)([array])


def get_blank():
    return tuple()


def concatenate(concater, panel, axis):

    if DEBUG_LEVEL > DEBUG_INFO:
        print("Concatenate IN (concat, panel, axis): ", concater, panel, axis)

    from pyslabs import slabif

    if concater is None:
        return panel

    if type(concater) != type(panel) and not isinstance(concater, ScalarList):
        raise PE_Stabif_Typemismatch("%s != %s" % (type(concater).__name__,
                                    type(panel).__name__))

    if isinstance(concater, ScalarList):
        concater.concat(panel)

    elif isinstance(concater, list):
        if axis > 0:
            if len(concater) != len(panel):
                raise PE_Stabif_Lengthmismatch("%d != %d" % (len(concater),
                                        len(panel)))
            buf = []
            for idx in range(len(concater)):
                buf.append(slabif.concatenate(concater[idx], panel[idx], axis=axis-1))
            concater = buf 

        else:
            concater.extend(panel)

    elif isinstance(concater, tuple):
        if axis > 0:
            if len(concater) != len(panel):
                raise PE_Stabif_Lengthmismatch("%d != %d" % (len(concater),
                                        len(panel)))
            buf = []
            for idx in range(len(concater)):
                buf.append(slabif.concatenate(concater[idx], panel[idx], axis=axis-1))
            concater = tuple(buf) 

        else:
            concater += panel

    else:
        if axis > 0:
            raise PE_Stabif_Wrongaxis()

        concater = ScalarList((concater, panel))


    if DEBUG_LEVEL > DEBUG_INFO:
        print("Concatenate OUT: ", concater)
    return concater



def squeeze(array):
    if DEBUG_LEVEL > DEBUG_INFO:
        print("Squeeze IN: ", array)

    is_squeezed = False

    if isinstance(array, (list, tuple, ScalarList)):
        if len(array) == 1:
            array = array[0]
            is_squeezed = True

    else:
        import pdb; pdb.set_trace()        

    if DEBUG_LEVEL > DEBUG_INFO:
        print("Squeeze OUT(array, is_squeezed): ", array, is_squeezed)

    return is_squeezed, array


def get_slice(slab, key):

    from pyslabs import slabif

    if not key:
        return slab

    if isinstance(key, (int, slice)):
        key = (key,)

    if isinstance(key[0], int):
        is_slice = False
        start, stop, step = key[0], key[0]+1, 1

    elif isinstance(key[0], slice):
        is_slice = True
        start, stop, step = key[0].start, key[0].stop, key[0].step

    buf = []

    for item in itertools.islice(slab, start, stop, step):
        if len(key) > 1:
            item_slice = slabif.get_slice(item, key[1:])
            buf.append(item_slice)
        else:
            buf.append(item)

    if not is_slice:
        return buf[0]

    else:
        return slice2type(buf, type(slab))



import os, pickle, resource, itertools
import pyslabs.data_numpy as dnp
from collections import OrderedDict
from tarfile import TarInfo

_supported_arrays = {
    "numpy": (lambda a: (type(a).__name__=="ndarray" and
                        type(a).__module__== "numpy"), "npy")
}

_cache = OrderedDict()


def _clear_cache():
    _cache.clear()


def _evict(num_items=1):
    _cache.popitem(last=False)


def arraytype(slab):
    for atype, (check, ext) in _supported_arrays.items():
        if check(slab):
            return atype, ext
        
    return "pickle", "dat"


def dump(slab, file):

    atype, ext = arraytype(slab)

    if atype == "numpy":
        return dnp.dump(slab, file)

    try:
        pickle.dump(slab, file)
        file.flush()
        os.fsync(file.fileno())

    except Exception as err:
        with open(file, "wb") as fp:
            pickle.dump(slab, fp)
            fp.flush()
            os.fsync(fp.fileno())


#def ccconcat(arrays, atype):
#
#    if not arrays:
#        return arrays
#
#    if atype == "numpy":
#        return dnp.concat(arrays, axis=0)
#
#    try:
#        output = arrays[0]
#        for item in arrays[1:]:
#            output += item
#
#    except TypeError:
#        output = arrays[0]
#        for item in arrays[1:]:
#            if isinstance(output, dict):
#                output.update(item)
#            else:
#                raise Exception("Not supported type for concatenation: %s" %
#                                str(type(output)))
#
#    return output


def stack(arrays, atype):

    if not arrays:
        return arrays

    if atype == "numpy":
        return dnp.stack(arrays)

    try:
        return type(arrays[0])(arrays)

    except TypeError:
        return arrays


def _get_slice(slab, key):
    
    if not key:
        return slab

    start, stop, step = key[0].start, key[0].stop, key[0].step

    _stype = type(slab)
    _m = []

    for item in itertools.islice(slab, start, stop, step):
        _m.append(_get_slice(item, key[1:]))

    if _m:
        try:
            _m = _stype(_m)

        except TypeError:
            pass

    return _m


def get_slice(slab, atype, key):

    if slab is None:
        return slab

    if atype == "numpy":
        return dnp.get_slice(slab, key)

    return _get_slice(slab, key)   


def load(tfile, slabobj, atype):

    path = slabobj.path

    if path in _cache:
        return _cache[path]

    # TODO: add caching evict logic
    # usage = resource.getrusage().ru_maxrss

    file = tfile.extractfile(slabobj)

    if atype == "numpy":
        _d = (atype, dnp.load(file))
        _cache[path] = _d
        return _d

    slab = pickle.load(file)

    _d = ("pickle", slab)
    _cache[path] = _d
    return _d


def shape(slab):

    atype, ext = arraytype(slab)

    if atype == "numpy":
        return dnp.shape(slab)

    s = []

    while slab:
        try:
            l = len(slab)

            if l > 0:
                s.append(l)
                slab = slab[0]

            else:
                break
        except TypeError:
            break

    return tuple(s)


def ndim(slab):

    atype, ext = arraytype(slab)

    if atype == "numpy":
        return dnp.ndim(slab)

    return(len(shape(slab)))


def length(slab, dim):
    _s = shape(slab)
    return _s[dim]


def squeeze_dim0(slab, atype):

    if atype == "numpy":
        return dnp.squeeze(slab)

    if length(slab, 0) == 1:
        return slab[0]


def _concat(bucket, array):

    if bucket[1] is None:
        bucket[0] = array[0]
        bucket[1] = array[1]
        return

    if bucket[0] == array[0] or bucket[0] is None:
        atype = array[0]

    elif array[0] is None:
        atype = bucket[0]

    else:
        import pdb; pdb.set_trace()

    if atype == "numpy":
        bucket[0] = atype
        bucket[1] = dnp.concat((bucket[1], array[1]), axis=1)
        return

    bucket[0] = atype

    for i, item in enumerate(array[1]):
        bucket[1][i] = bucket[1][i] + array[1][i]


def _merge(tfile, slabobj):

    _d = []
    _f = []
    _atype = None


    sortedkey = sorted([(int(k.split(".")[0]), k) for k in slabobj], key=lambda x: x[0])

    for (_, key) in sortedkey:
        item = slabobj[key]

        if isinstance(item, dict):
            _d.append(_merge(tfile, item))

        elif isinstance(item, TarInfo):
            _, atype, _ = os.path.basename(item.path).split(".")

            if _atype is None:
                _atype = atype
                # TODO: check size  of loaded array based on key and next key
                _f.append(load(tfile, item, atype)[1])

            elif _atype != atype:
                raise Exception("Different type exists in a stack: %s != %s" % (_atype, atype))

            else:
                _f.append(load(tfile, item, atype)[1])
            
        else:
            raise Exception("Unknown file type: %s" % str(item))

    if _f:
        _m = [_atype, stack(_f, atype)]

    else:
        _m = [None, None]

    for _i in _d:
        _concat(_m, _i)
        
    if _m:
        try:
            _m[1] = type(_m[1][0])(_m[1])

        except TypeError:
            pass

    return _m

def get_array(tfile, slabobj, pack_stack_dim):

    stype, arr = _merge(tfile, slabobj)

    try:
        if pack_stack_dim and arr is not None and length(arr, 0) == 1:
            arr = squeeze_dim0(arr, stype)
    except IndexError as err:
        pass

    return arr



import numpy as np
from io import BytesIO

def dump(slab, file):
    return np.save(file, slab)


def concat(arrays, axis=0):
    try:
        return np.concatenate(arrays, axis=axis)
    except Exception as err:
        import pdb; pdb.set_trace()
        print(err)


def stack(arrays):
    return np.stack(arrays)


def load(file):
    bytes = BytesIO()
    bytes.write(file.read())
    bytes.seek(0)
    return np.load(bytes)


def shape(slab):
    return slab.shape


def ndim(slab):
    return slab.ndim


def length(slab, axis):
    _s = shape(slab)
    return _s[axis]


def squeeze(slab):
    return np.squeeze(slab, axis=0)


def get_slice(slab, key):
    return slab.__getitem__(tuple(key))



"""Pyslabs utility module

"""

import os, io, pickle, shutil

from pyslabs.error import PE_Util_Typemismatch


supported_array_types = {
    "numpy": (lambda a: (type(a).__name__=="ndarray" and
                        type(a).__module__== "numpy"), "npy")
}

DEBUG_NONE, DEBUG_MAJOR, DEBUG_MINOR, DEBUG_INFO, DEBUG_ALL = range(5)
DEBUG_LEVEL = DEBUG_MAJOR
#DEBUG_LEVEL = DEBUG_ALL

class ScalarList():

    def __init__(self, elems=None, type=type(None)):

        self._list = list(elems)
        self._type = type

    def concat(self, elem):

        self._list.append(elem)

        if self._type is type(None):
            self._type = type(elem)

        elif self._type is not type(elem):
            raise PE_Util_Typemismatch("%s != %s" % (self._type, type(elem)))


def arraytype(slab):
    for atype, (check, ext) in supported_array_types.items():
        if check(slab):
            return atype, ext

    return "pickle", "dat"


def pickle_dump(path, obj):
    with io.open(path, "wb") as fp:
        pickle.dump(obj, fp)
        fp.flush()
        os.fsync(fp.fileno())


def clean_folder(folder):
    for files in os.listdir(folder):
        path = os.path.join(folder, files)
        try:
            shutil.rmtree(path)
        except OSError:
            os.remove(path)



import os, sys, io, pickle, shutil, time, uuid, tarfile, copy, itertools, pprint
from collections import OrderedDict
from pyslabs import data


UNLIMITED = -1
_CONFIG_FILE = "__config__"
_BEGIN_FILE = "__begin__"
_VARCFG_FILE = "__varcfg__"
_FINISHED = "__finished__"
_EXT = ".slab"
_CEXT = ".zlab"
_BEGIN_EXT = ".__slabbegin__"
_WORKDIR_EXT = ".__slabtmp__"
_MAX_OPEN_WAIT = 10 # seconds
_MAX_CLOSE_WAIT = 100 # seconds
_CONFIG_INIT = {
    "version": 1,
    "dims": {},
    "vars": {},
    "attrs": {},
    "__control__": {
        "nprocs": 1,
    }
} 
_VARCFG_INIT = {
    "writes": {},
    "shape": None,
    "check": {}
} 
_DIMCFG_INIT = {
} 

class Dimension():

    def __init__(self, config):
        self.config = config

    def __getattr__(self, name):
        if name in self.config:
            return self.config[name]

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, d):
        self.__dict__.update(d)

    def check(self, name, that):

        this = getattr(self, name, None)

        if name == "length":
            if this != UNLIMITED and that != this:
                raise Exception("stack dimension mismatch: %d != %d" %
                        (that, this))
        else:
            if that != this:
                raise Exception("dimension attribute mismatch: %s != %s" %
                        (str(that), str(this)))


class StackDimension(Dimension):
    pass


class VariableWriter():

    def __init__(self, path, config):

        self.path = path
        self.config = config
        self.writecount = 0

    def write(self, slab, start=None, shape=None):

        # get slab info
        slabshape = data.shape(slab)
        slabndim = len(slabshape)

        # shape check
        if shape:
            if tuple(shape) != slabshape:
                raise Exception("Shape check fails: %s != %s" %
                        (str(shape) != str(slabshape)))

        if start is None:
            start = (0,) * len(slabshape)

        elif isinstance(start, int):
            start = (start,) + (0,) * (len(slabshape) - 1)

        else:
            start = start + (0,) * (len(slabshape) - len(start))

        # generate shape
        if self.config["shape"] is None:
            self.config["shape"] = [1] + list(slabshape)

        else:
            if tuple(self.config["shape"][1:]) != tuple(slabshape):
                raise Exception("Shape check fails: %s != %s" %
                        (str(self.config["shape"][1:]) != str(slabshape)))

            self.config["shape"][0] += 1

        # generate relative path to data file

        slabpath = []

        for _s in start:
            slabpath.append(str(_s))

        wc = str(self.writecount)
        
        if wc in self.config["writes"]:
            writes = self.config["writes"][wc]

        else:
            writes = {}
            self.config["writes"][wc] = writes

        writes["/".join(slabpath)] = (start, slabshape)

        path = os.path.join(self.path, *slabpath)

        if not os.path.isdir(path):
            os.makedirs(path)

        atype, ext = data.arraytype(slab)
        slabpath = os.path.join(path, ".".join([wc, atype, ext])) 

        data.dump(slab, slabpath)

        self.writecount += 1


# TODO: add slab-level services
# TODO: use buffer for tarfile

class VariableReader():

    def __init__(self, tfile, slabtower, config, dims, pack_stack_dim, start=None):

        self._tfile = tfile
        self._slabtower = slabtower
        self._dims = dims
        self._config = config
        self._pack_stack_dim = pack_stack_dim
        self.shape = tuple(self._config["shape"]) # can be dim names or int values

#    @property
#    def ndim(self):
#
#        if self.shape[0] == 1:
#            return len(self.shape) - 1
#
#        else:
#            return len(self.shape)

#    def __len__(self):
#        s = self.shape[0]
#        return (self._dims[s]["length"] if s in self._dims else s)

    def _get_slice(self, k, length):

        if isinstance(k, int):
            if k < 0:
                start, stop, step = length+k, length+k+1, 1

            else:
                start, stop, step = k, k+1, 1

        else:

            start = 0 if k.start is None else k.start
            stop = length if k.stop is None else k.stop
            step = 1 if k.step is None else k.step

            if step < 0:
                raise Exception("Slicing step should be positve integer: %s" % str(k))

            if start < 0:
                start = length + start

            if stop < 0:
                stop = length + stop

        return slice(start, stop, step)

#
#        minval = length
#        maxval = 0
#
#        # TODO: REVISE THIS 
#        for i in range(start, stop, step):
#            if i < minval:
#                minval = i
#
#            if i > maxval:
#                maxval = i
#
#        return slice(minval, maxval + 1, step)

    def _merge_stack(self, tower, tkey, shape, newkey):

        #For a positive step, r[i] = start + step*i where i >= 0 and r[i] < stop.
        #For a negative step, r[i] = start + step*i, but the constraints are i >= 0 and r[i] > stop.

        #TODO: manipulate tkey (None?) to skip stack

        _k = self._get_slice(tkey, shape[0])

        _m = []
        atype = None

        try:
            for name, tinfo in itertools.islice(tower.items(),
                    _k.start, _k.stop, _k.step):
                _, _atype, _ = name.split(".")

                if atype is None:
                    atype = _atype

                elif atype != _atype:
                    raise Exception("Array type mismatch: %s != %s" % (str(atype), str(_atype)))

                _, slab = data.load(self._tfile, tinfo, atype)

                slab = data.get_slice(slab, atype, newkey)

                _m.append(slab)

        except ValueError as err:

            import pdb; pdb.set_trace()
            print(err)

        _a = data.stack(_m, atype)

        # _a shape shoud be the same to the original stack shape

        return (atype, _a)

    def _get_array(self, tower, key, tkey, shape, newkey=None):

        # create a new list for processing at this level
        newkey = list() if newkey is None else list(newkey)

        if len(key) == 0:
            return self._merge_stack(tower, tkey, shape, newkey)

        _k = self._get_slice(key[0], shape[0])

        _m = []

        # next indices
        nidxes = list(tower.keys())[1:] + [str(shape[0])]

        atype = None
        offset = 0

        for (idx, val), nidx in zip(tower.items(), nidxes):

            idx = int(idx)
            nidx = int(nidx)

            if nidx <= _k.start:
                continue

            if idx >= _k.stop:
                break

            if idx <= _k.start and _k.start < nidx:
                a = _k.start - idx

                if _k.stop >= nidx:
                    b = nidx - idx

                else:
                    b = _k.stop - idx

            elif idx <= _k.stop and _k.stop <= nidx:
                a = offset
                b = _k.stop - idx

            else:
                a = offset
                b = nidx - idx

            offset = (_k.step - (b - a) % _k.step) % _k.step

            _nk = newkey + [slice(a, b, _k.step)]

            #_atype, _o = self._get_array(val, key[1:], tkey, shape[1:], newkey)
            _atype, _o = self._get_array(val, key[1:], tkey, shape[1:], _nk)

            if _o is not None:
                if atype is None:
                    atype = _atype

                elif atype != _atype:
                    raise Exception("Array type mismatch: %s != %s" % (str(atype), str(_atype)))

                _m.append((atype, _o))

        _x = [None, None]

        for _i in _m:
            data._concat(_x, _i)

        import pdb; pdb.set_trace()
        return _x

    def __getitem__(self, key):

        ndim = len(self.shape)

        if self._pack_stack_dim and self.shape[0] == 1:
            key = (0, key) if isinstance(key, int) else ((0,) + key)

        if isinstance(key, int):
            key = (key,) + (slice(None, None, None),) * (ndim-1)

        elif len(key) < ndim:
            key = key + (slice(None, None, None),) * (ndim-len(key))


        for k in key:
            if isinstance(k, slice) and k.step is not None and k.step <= 0:
                raise Exception("Slice step should be positive integer: %s" % str(key))

        shape = []
        for s in self.shape:
            if s in self._dims:
                shape.append(self._dims[s]["length"])

            else:
                shape.append(s)

        # put stack dimension at the last
        shape = shape[1:] + [shape[0]]

        atype, array = self._get_array(self._slabtower, key[1:], key[0], shape)

        try:
            if self._pack_stack_dim and array is not None and data.length(array, 0) == 1:
                array = data.squeeze_dim0(array, atype)
        except IndexError as err:
            pass

        return array

    def __iter__(self):
        import pdb; pdb.set_trace()

    def __next__(self):
        import pdb; pdb.set_trace()

    def __contains__(self, item):
        import pdb; pdb.set_trace()


class PyslabsWriter():

    def __init__(self, root, config):
        self.root = root
        self.uuid = str(uuid.uuid4().hex)
        self.path = os.path.join(self.root, self.uuid)
        self.cfgpath = os.path.join(self.root, _CONFIG_FILE)
        self.config = config

        os.makedirs(self.path)

    def close(self):

        for name, cfg in self.config["vars"].items():
            with io.open(os.path.join(self.path, name, _VARCFG_FILE), "wb") as fp:
                pickle.dump(cfg, fp)
                fp.flush()
                os.fsync(fp.fileno())

        with io.open(os.path.join(self.path, _FINISHED), "w") as fp:
            fp.write("FINISHED")
            fp.flush()
            os.fsync(fp.fileno())


class ParallelPyslabsWriter(PyslabsWriter):

    def get_writer(self, name, **kwargs):

        varcfg = self.config["vars"][name]

        return VariableWriter(os.path.join(self.path, name), varcfg)
    
    def get_dim(self, name):

        dimcfg = self.config["dims"][name]

        return Dimension(dimcfg)

    def get_stack(self, name):

        dimcfg = self.config["dims"][name]

        return StackDimension(dimcfg)

    def begin(self):
        # place holder
        pass


class MasterPyslabsWriter(PyslabsWriter):

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        return self.close()

    def begin(self):

        with io.open(self.cfgpath, "wb") as fp:
            pickle.dump(self.config, fp)
            fp.flush()
            os.fsync(fp.fileno())

        procs = []
 
        start = time.time()
        nprocs = self.config["__control__"]["nprocs"]

        while time.time() - start < _MAX_OPEN_WAIT:

            procs.clear()

            for item in os.listdir(self.root):
                if item == self.uuid:
                    procs.append(os.path.join(self.root, item))
                    time.sleep(0.1)
                    continue

                try:
                    if len(item) == len(self.uuid) and int(item, 16):
                        proc = os.path.join(self.root, item)
                        procs.append(proc)

                except ValueError:
                    pass

            if len(procs) == nprocs:
                break

        if len(procs) != nprocs:
            raise Exception("Number of processes mismatch: %d != %d" %
                    (len(procs), nprocs))

    def get_writer(self, name, shape=None, **kwargs):

        varcfg = copy.deepcopy(_VARCFG_INIT)

        if shape is not None:
            if isinstance(shape, (int, StackDimension)):
                _shape = (shape if isinstance(shape, StackDimension) else
                            self.define_stack("stack", shape))

            else:
                _stack = (shape[0] if isinstance(shape[0], StackDimension) else
                            self.define_stack("stack", int(shape[0])))

                _shape = [_stack]
                for i, s in enumerate(shape[1:]):
                    _dim = (s if isinstance(s, Dimension) else
                        self.define_dim("dim%d"%(i+1), int(s)))
                    _shape.append(_dim)
 
            varcfg["check"]["shape"] = _shape

        self.config["vars"][name] = varcfg

        return VariableWriter(os.path.join(self.path, name), varcfg)

    def define_dim(self, name, length, origin=(0, "O"), unit=(1, ""),
                    points=None, desc="N/A", **kwargs):

        if isinstance(origin, int):
            origin = (origin, "O")

        if isinstance(unit, int):
            unit = (unit, "")

        if points is not None:
            _length = len(points)
            if _length != length:
                raise Exception("Dimension '%s' length mismatch: %d != %d" %
                                (name, length, _length))

            origin = (points[0], origin[1])
            unit = (None, unit[1])

        dimcfg = copy.deepcopy(_DIMCFG_INIT)
        dimcfg["name"] = name
        dimcfg["length"] = length
        dimcfg["origin"] = origin
        dimcfg["unit"] = unit
        dimcfg["points"] = points
        dimcfg["desc"] = desc
        dimcfg["attrs"] = dict((k[5:],v) for k,v in kwargs.items() if
                                k.startswith("attr_"))

        self.config["dims"][name] = dimcfg

        return Dimension(dimcfg)

    def define_stack(self, name, length, origin=(0, "O"), unit=(1, "slab"),
                    points=None, desc="N/A", **kwargs):

        dim = self.define_dim(name, length, origin=origin, unit=unit,
                                points=points, desc=desc, **kwargs)

        if dim.config["unit"]:
            if (not isinstance(dim.config["unit"][0], int) and
                dim.config["unit"][0] < 0):
                raise Exception("Stack unit is not a positive integer: %s" %
                    str(dim.config["unit"][0]))
        
        return StackDimension(dim.config)

    def close(self):
 
        super(MasterPyslabsWriter, self).close()

        beginpath = self.config["__control__"]["beginpath"]

        if os.path.isfile(beginpath):
            os.remove(beginpath)

        def _move_dim(src, dst, attrs):
            
            for dim in os.listdir(src):
                srcpath = os.path.join(src, dim)
                dstpath = os.path.join(dst, dim)

                if os.path.isdir(srcpath):
                    if os.path.isdir(dstpath):
                        _move_dim(srcpath, dstpath, attrs) 

                    elif os.path.exists(dstpath):
                        raise Exception("Destination path already exists: %s" % dstpath)

                    else:
                        shutil.move(srcpath, dstpath)

                elif os.path.exists(dstpath):
                    raise Exception("Multiple processes creat the same data file: %s" % dstpath)

                else:
                    shutil.move(srcpath, dstpath)
              
        def _move_proc(src, dst, attrs):

            for var in os.listdir(src): 

                dstvar = os.path.join(dst, var)
                srcvar = os.path.join(src, var)

                if not var.startswith("_") and var not in attrs["vars"]:
                    attrs["vars"][var] = {"config": []}

                varcfg = os.path.join(srcvar, _VARCFG_FILE)

                with io.open(varcfg, "rb") as fp:
                    _cfg = pickle.load(fp)

                attrs["vars"][var]["config"].append(_cfg)

                os.remove(varcfg)

                if os.path.isdir(dstvar):
                    _move_dim(srcvar, dstvar, attrs) 

                else:
                    shutil.move(srcvar, dstvar)

        procs = []

        start = time.time()
        nprocs = self.config["__control__"]["nprocs"]

        while time.time() - start < _MAX_CLOSE_WAIT:

            procs.clear()

            for item in os.listdir(self.root):
                if item == self.uuid:
                    procs.append(os.path.join(self.root, item))
                    time.sleep(0.1)
                    continue

                try:
                    if len(item) == len(self.uuid) and int(item, 16):
                        proc = os.path.join(self.root, item)
                        procs.append(proc)

                except ValueError:
                    pass

            if len(procs) == nprocs:
                break

        if len(procs) != nprocs:
            raise Exception("Number of processes mismatch: %d != %d" %
                    (len(procs), nprocs))

        for proc in procs:
            finished = os.path.join(proc, _FINISHED)
            timeout = True

            while time.time() - start < _MAX_CLOSE_WAIT:
                if os.path.isfile(finished):
                    os.remove(finished)
                    timeout = False
                    break
                time.sleep(0.1)

            if timeout:
                raise Exception("Error: timeout on waiting for parallel process finish.")

        attrs = {"vars": {}}

        # restructure data folders
        for src in procs:
            _move_proc(src, self.root, attrs)
            shutil.rmtree(src)

        _shape = {}

        for vn, vc in attrs["vars"].items(): # var name, var config
            _sp = None # temp shape

            for _vcfg in vc["config"]:
                _vshape = _vcfg["shape"]
                if _sp is None:
                    _sp = _vshape

                elif _sp[0] != _vshape[0] or _sp[2:] != _vshape[2:]:
                    raise Exception("Shape mismatch: %s != %s" % (str(_sp), str(_vshape)))

                else:
                    _sp[1] += _vshape[1]

            _shape[vn] = _sp


        for name, varcfg in self.config["vars"].items():

            varcfg["shape"] = _shape[name]
            varcfg.pop("writes")

            if varcfg["check"]:
                for check in varcfg["check"].keys():
                    test = varcfg["check"][check]

                    if check == "shape":
                        if isinstance(test, StackDimension):
                            test.check("length", _shape[name][0])
                            varcfg["shape"] = test.name

                        elif len(test) > 0:
                            test[0].check("length", _shape[name][0])
                            varcfg["shape"][0] = test[0].name

                            for i, (_test, _length) in enumerate(zip(test[1:], _shape[name][1:])):
                                _test.check("length", _length)
                                varcfg["shape"][i+1] = _test.name
                    else:
                        raise Exception("Unknown variable test: %s" % check)

                varcfg.pop("check")

        archive = self.config["__control__"]["archive"]
        slabpath = self.config["__control__"]["slabpath"]

        self.config.pop("__control__")

        with io.open(self.cfgpath, "wb") as fp:
            pickle.dump(self.config, fp)
            fp.flush()
            os.fsync(fp.fileno())

        # archive if requested
        if archive:
            dirname, basename = os.path.split(self.root)

            with tarfile.open(slabpath, "w") as tar:
                for item in os.listdir(self.root):
                    itempath = os.path.join(self.root, item)
                    tar.add(itempath, arcname=item)

            shutil.rmtree(self.root)


class ParallelPyslabsReader():

    def __init__(self, slabpath):
        self.slabpath = slabpath
        self.slabarchive = tarfile.open(slabpath, mode="r:")
        self.slabtower = OrderedDict()

        _tower = {}

        for entry in self.slabarchive:
            if entry.name == _CONFIG_FILE:
                self.config = pickle.load(self.slabarchive.extractfile(entry))
            else:
                self._trie(_tower, entry.path.split("/"), entry)

        self._sort_tower(self.slabtower, _tower)

    def _sort_tower(self, st, t):

        for k in sorted(t.keys()):
            v = t[k]

            if isinstance(v, dict):
                _st = OrderedDict()
                st[k] = _st
                self._sort_tower(_st, v)

            else:
                st[k] = v

    def _trie(self, tower, path, entry):

        if len(path) == 1:

            if isinstance(tower, tarfile.TarInfo):
                import pdb; pdb.set_trace()

            elif path[0] in tower:
                raise Exception("Wrong mapping: %s" % path)

            tower[path[0]] = entry

        elif path[0] in tower:

            if isinstance(tower[path[0]], tarfile.TarInfo):
                tower[path[0]] = {}
                self._trie(tower[path[0]], path[1:], entry)

            else:
                self._trie(tower[path[0]], path[1:], entry)
        else:
            _t = {}
            tower[path[0]] = _t
            self._trie(_t, path[1:], entry)

    def get_reader(self, name, pack_stack_dim=False):

        varcfg = self.config["vars"][name]
        dimcfg = self.config["dims"]

        return VariableReader(self.slabarchive, self.slabtower[name], varcfg, dimcfg, pack_stack_dim)

    def get_array(self, name, pack_stack_dim=False):

        return data.get_array(self.slabarchive, self.slabtower[name], pack_stack_dim)

    def close(self):
        self.slabarchive.close()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def _traverse(self, tree, bag):
            try:
                for k, v in tree.items():
                    if bag["check"](k, v):
                        bag["data"].append(bag["output"](k, v))
                    self._traverse(v, bag)

            except AttributeError as err:
                pass

            except Exception as err:
                import pdb; pdb.set_trace()
                print(err)

    def info(self, mode, *args, **kwargs):

        if mode == "list":
            return tuple(self.slabtower.keys())

        elif mode == "var":
            out = {}
            var = self.get_reader(args[0])

            import pdb; pdb.set_trace()

        elif mode == "slab":
            out = {}
            bag = {"check": lambda k, v: isinstance(v, tarfile.TarInfo),
                    "output": lambda k, v: v}

            for var, tree in self.slabtower.items():
                data = []
                bag["data"] = data
                self._traverse(tree, bag)

                if len(data) > 0:
                    totalsize = 0
                    maxsize = 0
                    minsize = sys.maxsize
                    nslabs = len(data)

                    for tar in data:
                        totalsize += tar.size
                        if tar.size > maxsize:
                            maxsize = tar.size
                        if tar.size < minsize:
                            minsize = tar.size
                else:
                    totalsize = maxsize = minsize = nslabs = 0

                out[var] = (nslabs, totalsize, maxsize, minsize)

            return out

        elif mode == "":

            out = []

            out.append(("version", self.config["version"]))

            dbuf = []
            for n, d in self.config["dims"].items():
                dbuf.append((n,  d["length"]))

            out.append(("dims", tuple(dbuf)))

            vbuf = []
            for n, v in self.config["vars"].items():
                if "shape" in v:
                    vbuf.append((n, v["shape"]))
                else:
                    vbuf.append((n, None))

            out.append(("vars", tuple(vbuf)))
            out.append(("size", os.path.getsize(self.slabpath)))

            return out


class MasterPyslabsReader(ParallelPyslabsReader):
    pass


def master_open(slabpath, nprocs, mode="r", archive=True, workdir=None):

    if mode == "r" and nprocs > 1:
        print("Open with 'r' mode does not support parallel processing.")
        sys.exit(-1)

    data._clear_cache()

    if mode == "w":

        if slabpath.endswith(_EXT) or slabpath.endswith(_CEXT):
            base, ext = os.path.splitext(slabpath)
            beginpath = base + _BEGIN_EXT

            if workdir is None:
                workdir = base + _WORKDIR_EXT
        else:
            beginpath = slabpath + _BEGIN_EXT
            if workdir is None:
                workdir = slabpath + _WORKDIR_EXT
            slabpath += _EXT

        with io.open(beginpath, "wb") as fp:
            begin = {"workdir": workdir, "slabpath": slabpath, "mode":mode}
            pickle.dump(begin, fp)
            fp.flush()
            os.fsync(fp.fileno())

        # create root directory
        os.makedirs(workdir, exist_ok=True)
        for item in os.listdir(workdir):
            itempath = os.path.join(workdir, item)

            if os.path.isdir(itempath):
                shutil.rmtree(itempath)
            else:
                os.remove(itempath)

        if not os.path.isfile(beginpath):
            raise Exception("Can not create a flag file: %s" % beginpath)

        if not os.path.isdir(workdir):
            raise Exception("Work directory does not exist: %s" % workdir)

        cfg = copy.deepcopy(_CONFIG_INIT)
        cfg["__control__"]["nprocs"] = nprocs
        cfg["__control__"]["archive"] = archive
        cfg["__control__"]["beginpath"] = beginpath
        cfg["__control__"]["slabpath"] = slabpath

        return MasterPyslabsWriter(workdir, cfg)

    elif mode[0] == "r":

        if not os.path.isfile(slabpath):
            raise Exception("Slabfile is not found: %s" % slabpath)

        return MasterPyslabsReader(slabpath)

    else:
        raise Exception("Unknown open mode: %s" % str(mode))


def parallel_open(slabpath, mode="w"):

    data._clear_cache()

    if mode == "w":

        base, ext = os.path.splitext(slabpath)
        beginpath = (base if ext else slabpath) + _BEGIN_EXT

        start = time.time()
        while time.time() - start < _MAX_OPEN_WAIT:
            if os.path.isfile(beginpath):
                with io.open(beginpath, "rb") as fp:
                    begin = pickle.load(fp)
                    mode = begin["mode"]
                    workdir = begin["workdir"]
                break
            time.sleep(0.1)

        if workdir is None:
            raise Exception("No begin notification: %s" % beginpath)

        start = time.time()
        while time.time() - start < _MAX_OPEN_WAIT:

            cfgpath = os.path.join(workdir, _CONFIG_FILE)

            if not os.path.isfile(cfgpath):
                time.sleep(0.1)
                continue

            with io.open(cfgpath, "rb") as fp:
                cfg = pickle.load(fp)

            return ParallelPyslabsWriter(workdir, cfg)

    elif mode[0] == "r":

        return ParallelPyslabsReader(slabpath)

    else:
        raise Exception("Unknown open mode: %s" % str(mode))

    raise Exception("Target configuration is not configured: %s" % cfgpath)


def open(outfile, *vargs, **kwargs):

    nprocs = kwargs.pop("num_procs", 1)

    if len(vargs) > 0:
        if "mode" in kwargs:
            if vargs[0] != kwargs["mode"]:
                raise Exception("open mode mismatch")

            vargs = []

        else:
            kwargs["mode"] = vargs[0]

    return master_open(outfile, nprocs, **kwargs)



"""Pyslabs slab manipulation module


"""

import os, io, pickle, itertools, pprint

from pyslabs.util import arraytype, DEBUG_LEVEL, DEBUG_INFO, DEBUG_MAJOR
from collections import OrderedDict
import pyslabs.slabif_numpy as npif
import pyslabs.slabif_builtins as bif

_cache = OrderedDict()


def length(slab, axis=0):

    if slab is None:
        return 0

    atype, ext = arraytype(slab)

    if atype == "numpy":
        l = npif.length(slab, axis=axis)

    else:
        l = bif.length(slab, axis=axis)

    return l


def shape(slab):

    if slab is None:
        return tuple()

    atype, ext = arraytype(slab)

    if atype == "numpy":
        sp = npif.shape(slab)

    else:
        sp = bif.shape(slab)

    return sp


def ndim(slab):

    if slab is None:
        return 0

    return len(shape(slab))


def dump(path, slab):
    if DEBUG_LEVEL > DEBUG_INFO:
        print("Slabif dump IN (path, slab): ", path, slab)

    atype, ext = arraytype(slab)

    if atype == "numpy":
        out = npif.dump(path, slab)

    else:
        out = bif.dump(path, slab)

    return out


def load(tar_file, slab_info, atype):

    path = slab_info.path

    if path in _cache:
        return _cache[path]

    tar_file = tar_file.extractfile(slab_info)

    if atype == "numpy":
        slab = npif.load(tar_file)

    else:
        slab = bif.load(tar_file)

    _cache[path] = slab

    return _cache[path]


def expand_dim(slab):

    if slab is None:
        return slab

    atype, ext = arraytype(slab)

    if atype == "numpy":
        array = npif.expand_dim(slab)

    else:
        array = bif.expand_dim(slab)

    return array


def stack(upper, lower):

    atype, ext = arraytype(lower)

    try:
        if atype == "numpy":
            array = npif.stack(upper, lower)

        else:
            array = bif.stack(upper, lower)
    except Exception as err:
        import pdb; pdb.set_trace()
        if DEBUG_LEVEL > DEBUG_MAJOR:
            print(err)

    return array


def concatenate(concater, panel, axis):

    atype, ext = arraytype(concater)

    if atype == "numpy":
        array = npif.concatenate(concater, panel, axis)

    else:
        array = bif.concatenate(concater, panel, axis)

    return array

def squeeze(array):

    atype, ext = arraytype(array)

    if atype == "numpy":
        is_sq, sq= npif.squeeze(array)

    else:
        is_sq, sq = bif.squeeze(array)

    return is_sq, sq


# slice of array
def get_slice(array, key):

    if array is None:
        return slab

    atype, ext = arraytype(array)

    if atype == "numpy":
        sl = npif.get_slice(array, key)

    else:
        sl = bif.get_slice(array, key)

    return sl


def get_blank(atype):

    if atype == "numpy":
        bl = npif.get_blank()

    else:
        bl = bif.get_blank()

    return bl


def get_column(tar_file, slab_tower, stack_key, slab_key):


    if DEBUG_LEVEL > DEBUG_INFO:
        print("Get_Column IN (slat_tower, stack_key, slab_key): ", slab_tower.keys(), stack_key, slab_key)

    #For a positive step, r[i] = start + step*i where i >= 0 and r[i] < stop.
    #For a negative step, r[i] = start + step*i, but the constraints are i >= 0 and r[i] > stop.

    slab_type = None
    keys = sorted(slab_tower.keys(), key=lambda x:int(x.split(".")[0]))

    if isinstance(stack_key, int):
        is_slice = False
        stack_slice = slice(stack_key, stack_key+1, 1)

    else:
        is_slice = True
        stack_slice = stack_key

    if isinstance(stack_key, int):
        is_slice = False
        k0 = stack_key
        if k0 < 0:
            k0st = len(keys) + k0
            stack_slice = slice(k0st, k0st+1, 1)

        else:
            stack_slice = slice(k0, k0+1, 1)

    else:
        is_slice = True
        st = stack_key.start
        if st < 0: st = st + len(keys)
        so = stack_key.stop
        if so < 0: so = so + len(keys)
        se = stack_key.step
        if se < 0:
            raise PE_Slabif_Negativestep()

        stack_slice = slice(st, so, se)

    stacker = None

    if DEBUG_LEVEL > DEBUG_INFO:
        print("Gen_Column itertool islice: ", stack_slice)

    for key in itertools.islice(keys, stack_slice.start, stack_slice.stop,
                                stack_slice.step):
        tinfo = slab_tower[key]
        _, _stype, _ = key.split(".")

        if slab_type is None:
            slab_type = _stype

        elif slab_type != _stype:
            raise PE_Read_Slabtypemismatch("%s != %s" % (slab_type, _stype))

        slab_slice = get_slice(load(tar_file, tinfo, slab_type), slab_key)
        stacker = stack(stacker, slab_slice)

    #if not is_slice:
    #    stacker = squeeze(stacker)

    if DEBUG_LEVEL > DEBUG_INFO:
        print("Get_Column Out (squeezed, stacker): ", False, stacker)
    return False, stacker

def get_array(tar_file, slab_tower, slab_shape, slab_key, stack_key, new_key=None):
    if DEBUG_LEVEL > DEBUG_INFO:
        print("\nGet_array IN(tower, slab_shape, slab_key, stack_key, new_key): ", slab_tower.keys(), slab_shape, slab_key, stack_key, new_key)

    if len(slab_key) == 0:
        is_squeezed, column =  get_column(tar_file, slab_tower, stack_key, new_key)
        if DEBUG_LEVEL > DEBUG_INFO:
            print("Get_array Column: \n")
            pprint.pprint(column)

        return is_squeezed, column

    cidxes = []
    nidxes = []

    skeys = [k.split("_") for k in slab_tower.keys()]
    ikeys = sorted([(int(s), int(l)) for s, l in skeys], key=lambda x:x[0])
    for s, l in ikeys:
        cidxes.append(s)
        nidxes.append(s+l)
    assert nidxes[-1] == slab_shape[0]

    if isinstance(slab_key[0], int):
        is_slice = False
        k0 = slab_key[0]
        if k0 < 0:
            k0st = slab_shape[0] + k0
            ckey = slice(k0st, k0st+1, 1)

        else:
            ckey = slice(slab_key[0], slab_key[0]+1, 1)

    else:
        is_slice = True
        st = slab_key[0].start
        if st < 0: st = st + slab_shape[0]
        so = slab_key[0].stop
        if so < 0: so = so + slab_shape[0]
        se = slab_key[0].step
        if se < 0:
            raise PE_Slabif_Negativestep()

        ckey = slice(st, so, se)

    concater = None

    atype = None
    offset = 0

    #import pdb; pdb.set_trace()
    for cidx, nidx in zip(cidxes, nidxes):

        sub_tower = slab_tower[str(cidx)+"_"+str(nidx-cidx)]

        if nidx <= ckey.start:
            continue

        if cidx >= ckey.stop:
            break

        if cidx <= ckey.start and ckey.start < nidx:
            a = ckey.start - cidx

            if ckey.stop >= nidx:
                b = nidx - cidx

            else:
                b = ckey.stop - cidx

        elif cidx <= ckey.stop and ckey.stop <= nidx:
            a = offset
            b = ckey.stop - cidx

        else:
            a = offset
            b = nidx - cidx

        offset = (ckey.step - (b - a) % ckey.step) % ckey.step

        next_key = list() if new_key is None else list(new_key)

        if is_slice:
            last_key = slice(a, b, ckey.step)

        else:
            lslice = [s for s in range(a, b, ckey.step)]
            if len(lslice) == 1:
                last_key = lslice[0]

            else:
                raise PE_Slabif_Wrongslicing()

        next_key.append(last_key)

        is_squeezed, panel = get_array(tar_file, sub_tower, slab_shape[1:],
                                slab_key[1:], stack_key, next_key)
        if concater is None:
            concater = panel

        else:
            axis = (len(next_key)-1) if is_squeezed else len(next_key)
            if DEBUG_LEVEL > DEBUG_INFO:
                print("Slab Key, next_key, axis: ", slab_key[1:], next_key, axis)
            concater = concatenate(concater, panel, axis)

    is_squeezed_final = False

    if not is_slice:
        if DEBUG_LEVEL > DEBUG_INFO:
            print("TYPE: ", type(concater).__name__)
        if concater is None: import pdb; pdb.set_trace()
        is_squeezed_final, concater = squeeze(concater)

    if concater is None:
        concater = get_blank(atype)

    if DEBUG_LEVEL > DEBUG_INFO:
        print("Get_array OUT: ", concater)
    return is_squeezed_final, concater



"""main entry for pyslabs command-line interface"""

import pyslabs

def cmd_info(args):

    with pyslabs.open(args.slabfile) as fp:

        if args.list:
            print(", ".join(fp.info("list", verbose=args.verbose)))
  
        elif args.slab:
            print("") 
            for var, (n, t, x, m) in fp.info("slab", args.var, verbose=args.verbose).items():
                print(var) 
                print("   slab count: %d" % n) 
                print("  total bytes: %d" % t) 
                print("    max bytes: %d" % x) 
                print("    min bytes: %d" % m) 

        elif args.var:
            var = fp.info("var", args.var, verbose=args.verbose)

        else:
            for k, v in fp.info("", args.var, verbose=args.verbose):
                if k == "size":
                    print("size: %d bytes" % v)

                elif k == "dims":
                    buf = []
                    for n, l in v:
                        buf.append("%s(%d)" % (n, l))
                    print("dims: " + ", ".join(buf))

                elif k == "vars":
                    buf = []
                    for n, s in v:
                        buf.append("%s(%s)" % (n, ", ".join([str(_s) for _s in s])))
                    print("vars: " + ", ".join(buf))
                   

def main():
    import argparse
    from pyslabs.info import version

    parser = argparse.ArgumentParser(description="pyslabs command-line tool")
    parser.add_argument("--version", action="version", version="pyslabs "+version)
    parser.add_argument("--verbose", action="store_true", help="verbose info")

    cmds = parser.add_subparsers(title='subcommands',
                description='pyslabs subcommands', help='additional help')

    p_info = cmds.add_parser('info')
    p_info.add_argument("slabfile", help="slabfile path")
    p_info.add_argument("-l", "--list", action="store_true", help="a list of variables")
    p_info.add_argument("-v", "--var", help="variable info")
    p_info.add_argument("-s", "--slab", action="store_true", help="slab info")
    p_info.set_defaults(func=cmd_info)

    argps = parser.parse_args()
    argps.func(argps)

    return 0


if __name__ == "__main__":
    import sys
    sys.exit(main())



"pyslabs setup module"

def main():

    from setuptools import setup, find_packages
    from pyslabs.const import name, version, description, long_description, author

    console_scripts = ["slabs=pyslabs.command:main"]
    keywords = ["Parallel I/O", "pyslabs"]

    setup(
        name=name,
        version=version,
        description=description,
        long_description=long_description,
        author=author,
        author_email="youngsung.kim.act2@gmail.com",
        classifiers=[
            "Development Status :: 3 - Alpha",
            "Intended Audience :: Science/Research",
            "Topic :: Scientific/Engineering",
            "License :: OSI Approved :: MIT License",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.5",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
        ],
        keywords=keywords,
        packages=find_packages(exclude=["tests"]),
        include_package_data=True,
        entry_points={ "console_scripts": console_scripts },
        project_urls={
            "Bug Reports": "https://github.com/grnydawn/pyslabs/issues",
            "Source": "https://github.com/grnydawn/pyslabs",
        }
    )

if __name__ == '__main__':
    import multiprocessing
    multiprocessing.freeze_support()
    main()



"""Pyslabs core module


naming conventions
======================

path : may includes directory path
name : does not include directory path

"""

import os, sys, io, copy, time, uuid, pickle, shutil, tarfile

from collections import OrderedDict
from pyslabs.const import (SLAB_EXT, ZLAB_EXT, TMP_BEGIN, TMP_WORK, INIT_BEGIN,
                           INIT_CONFIG, INIT_VARCFG, INIT_DIMCFG, CONFIG_FILE,
                           FINISH_FILE, INIT_TIMEOUT, FINI_TIMEOUT, VARCFG_FILE,
                           UNLIMITED)
from pyslabs.error import PE_Begin_Numproc, PE_Close_Startindexerror, PE_Close_Shapemismatch
from pyslabs.util import pickle_dump, clean_folder
from pyslabs.write import VariableWriterV1
from pyslabs.read import VariableReaderV1
from pyslabs.slabif import _cache


##############################
# PRIVATE FUNCTIONS
##############################

def _write_paths(slab_path, work_path):

    begin_path = slab_path

    if slab_path.endswith(SLAB_EXT) or slab_path.endswith(ZLAB_EXT):
        base, ext = os.path.splitext(slab_path)

        if work_path is None:
            work_path = base + TMP_WORK

    else:

        if work_path is None:
            work_path = base + TMP_WORK

        slab_path += SLAB_EXT

#
#    if slab_path.endswith(SLAB_EXT) or slab_path.endswith(ZLAB_EXT):
#        base, ext = os.path.splitext(slab_path)
#
#        if work_path is None:
#            begin_path = base + TMP_BEGIN
#            work_path = base + TMP_WORK
#
#        else:
#            begin_path = work_path + TMP_BEGIN
#
#    else:
#
#        if work_path is None:
#            begin_path = slab_path + TMP_BEGIN
#            work_path = base + TMP_WORK
#
#        else:
#            begin_path = work_path + TMP_BEGIN
#
#        slab_path += SLAB_EXT

    return slab_path, begin_path, work_path


##############################
# PUBLIC CLASSES
##############################

class Dimension():

    def __init__(self, config):
        self.config = config

    def __getattr__(self, name):
        if name in self.config:
            return self.config[name]

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, d):
        self.__dict__.update(d)


    def check(self, length):

        mylen = self.config["length"]

        if mylen is None:
            return True

        if length <  0 or (mylen != UNLIMITED and mylen != length):
            return False

        return True

class StackDimension(Dimension):
    pass


# implement slab structure and protocol
class PyslabsWriterV1(object):

    def __init__(self, work_path, config):
        self.work_path = work_path
        self.uuid = str(uuid.uuid4().hex)
        self.proc_path = os.path.join(work_path, self.uuid)
        self.cfg_path = os.path.join(work_path, CONFIG_FILE)
        self.config = config

        os.makedirs(self.proc_path)

    def close(self):

        for name, cfg in self.config["vars"].items():
            cfg_path = os.path.join(self.proc_path, name, VARCFG_FILE)
            pickle_dump(cfg_path, cfg)

        finish_path = os.path.join(self.proc_path, FINISH_FILE)

        with io.open(finish_path, "w") as fp:
            fp.write("FINISHED")
            fp.flush()
            os.fsync(fp.fileno())


# master implementation of pyslabs 
class MasterPyslabsWriterV1(PyslabsWriterV1):

    def get_writer(self, name, shape=None, autostack=False, **kwargs):

        var_cfg = copy.deepcopy(INIT_VARCFG)

        if shape is not None:
            sh = []
            for i, l in enumerate(shape):
                if isinstance(l, Dimension):
                    sh.append(l)

                elif isinstance(l, int):
                    if i == 0:
                        self.define_stack("stack", l)
                    else:
                        self.define_dim("dim%d"%i, l)
 
                elif l is None:
                    if i == 0:
                        self.define_stack("stack", UNLIMITED)
                    else:
                        self.define_dim("dim%d"%i, UNLIMITED)
               
                else:
                    raise PE_Core_Invalidshapetype(str(shape))

            shape = tuple(shape)

        var_cfg["check"]["shape"] = shape
        var_cfg["stack"]["auto"] = autostack
        var_cfg["attrs"].update(dict((k[5:],v) for k,v in kwargs.items() if
                                k.startswith("attr_")))

        self.config["vars"][name] = var_cfg

        return VariableWriterV1(os.path.join(self.proc_path, name), var_cfg)

    def define_dim(self, name, length, origin=(0, "O"), unit=(1, ""),
                    points=None, desc="N/A", **kwargs):

        if isinstance(origin, int):
            origin = (origin, "O")

        if isinstance(unit, int):
            unit = (unit, "")

        if points is not None:
            _length = len(points)
            if length is not None and _length != length:
                raise Exception("Dimension '%s' length mismatch: %d != %d" %
                                (name, length, _length))

            origin = (points[0], origin[1])
            unit = (None, unit[1])

        dim_cfg = copy.deepcopy(INIT_DIMCFG)
        dim_cfg["name"] = name
        dim_cfg["length"] = length
        dim_cfg["origin"] = origin
        dim_cfg["unit"] = unit
        dim_cfg["points"] = points
        dim_cfg["desc"] = desc
        dim_cfg["attrs"] = dict((k[5:],v) for k,v in kwargs.items() if
                                k.startswith("attr_"))

        self.config["dims"][name] = dim_cfg

        return Dimension(dim_cfg)

    def define_stack(self, name, length, origin=(0, "O"), unit=(1, "slab"),
                    points=None, desc="N/A", **kwargs):

        dim = self.define_dim(name, length, origin=origin, unit=unit,
                                points=points, desc=desc, **kwargs)

        if dim.config["unit"]:
            if (not isinstance(dim.config["unit"][0], int) and
                dim.config["unit"][0] < 0):
                raise Exception("Stack unit is not a positive integer: %s" %
                    str(dim.config["unit"][0]))

        return StackDimension(dim.config)

    def begin(self):

        pickle_dump(self.cfg_path, self.config)

        procs = []

        start = time.time()
        num_procs = self.config["_control_"]["num_procs"]

        while time.time() - start < INIT_TIMEOUT:

            procs.clear()

            for item in os.listdir(self.work_path):
                if item == self.uuid:
                    procs.append(os.path.join(self.work_path, item))
                    time.sleep(0.1)
                    continue

                try:
                    if len(item) == len(self.uuid) and int(item, 16):
                        proc = os.path.join(self.work_path, item)
                        procs.append(proc)

                except ValueError:
                    pass

            if len(procs) == num_procs:
                break

        begin_path = self.config["_control_"]["begin_path"]

        if os.path.isfile(begin_path):
            os.remove(begin_path)

        if len(procs) != num_procs:
            raise PE_Begin_Numproc("%d != %d" %(len(procs), num_procs))

    def close(self):

        super(MasterPyslabsWriterV1, self).close()

        # dim: dimension to scan, start indices of the dimension, slab_shape
        # TODO : get shape info from var config of each procs
        #def _scan(dim, start, slab_shape):
        def _scan(dim, start_length):

            st = None
            sh = None
            prev_st_len = None

            # sorted start indices
            for st_len in sorted(start_length.keys(), key=lambda x:x[0]):

                # indices of next dimension
                next_dim = start_length[st_len]

                # if stack dimension
                if next_dim is None:

                    # return start index and # of slabs
                    return [st_len[0]], [st_len[1]]

                else:
                    # if non-stack dimension, go to next dimension
                    _st, _sh = _scan(dim+1, next_dim)
                    # return with start and shape

                    # if the first index
                    if st is None:
                        st = [st_len[0]] + _st
                    elif sum(prev_st_len) != sh[0]:
                        raise PE_Close_Startindexerror("%s != %d" % (prev_st_len, sh[0]))

                    if sh is None:
                        sh = [st_len[1]] + _sh
                    else:
                        sh[0] = sum(st_len)

                prev_st_len = st_len

            return st, sh

        def _move_dim(src, dst, start_length):

            nslabs = None

            for idx_len in os.listdir(src):

                src_path = os.path.join(src, idx_len)
                dst_path = os.path.join(dst, idx_len)

                if os.path.isdir(src_path):

                    st_len = tuple(int(i) for i in idx_len.split("_"))

                    if st_len not in start_length:
                        dim_st_len = {}
                        start_length[st_len] = dim_st_len

                    else:
                        dim_st_len = start_length[st_len]

                    if os.path.isdir(dst_path):
                        _move_dim(src_path, dst_path, dim_st_len)

                    elif os.path.exists(dst_path):
                        raise PE_Close_DestExist(dst_path)

                    else:
                        os.makedirs(dst_path)
                        _move_dim(src_path, dst_path, dim_st_len)
                        #shutil.move(src_path, dst_path)

                elif os.path.exists(dst_path):
                    raise PE_Close_DestDupulicated(dst_path)

                else:
                    nslabs = 1 if nslabs is None else nslabs + 1
                    shutil.move(src_path, dst_path)

            if nslabs is not None:
                start_length[(0, nslabs)] = None

        def _move_proc(src, dst, attrs):

            for var in os.listdir(src):

                dst_path = os.path.join(dst, var)
                src_path = os.path.join(src, var)

                if not var.startswith("_") and var not in attrs["vars"]:
                    start_length = {}
                    attrs["vars"][var] = {"config": [], "start_length":start_length}

                else:
                    start_length = attrs["vars"][var]["start_length"]

                cfg_path = os.path.join(src_path, VARCFG_FILE)

                with io.open(cfg_path, "rb") as fp:
                    cfg = pickle.load(fp)
                    attrs["vars"][var]["config"].append(cfg)

                os.remove(cfg_path)

                if not os.path.isdir(dst_path):
                    os.makedirs(dst_path)

                _move_dim(src_path, dst_path, start_length)

#        def _get_shape(writes):
#
#            height = 0
#            shape = None
#
#            for stack, write in writes.items():
#                _shape = [{}]*len(ndim)
#                for start, slab_shape in sorted(write.values(), key=lambda x:x[0]):
#                    _shape_merge(_shape, start, slab_shape)
#                    import pdb; pdb.set_trace()
#
#                if shape is None:
#                    shape = _shape
#
#                elif shape != _shape:
#                    raise PE_Close_Shapemismatch("%s != %s" %
#                            (str(shape), str(_shape)))
#
#                height += 1 
#
#            import pdb; pdb.set_trace()
#            return [height] + shape

        procs = []

        start_time = time.time()
        num_procs = self.config["_control_"]["num_procs"]

        while time.time() - start_time < FINI_TIMEOUT:

            procs.clear()

            for item in os.listdir(self.work_path):
                if item == self.uuid:
                    procs.append(os.path.join(self.work_path, item))
                    time.sleep(0.1)
                    continue

                try:
                    if len(item) == len(self.uuid) and int(item, 16):
                        proc = os.path.join(self.work_path, item)
                        procs.append(proc)

                except ValueError:
                    pass

            if len(procs) == num_procs:
                break

        if len(procs) != num_procs:
            raise PE_Close_Numproc("%d != %d" %(len(procs), num_procs))

        for proc in procs:
            finish_path = os.path.join(proc, FINISH_FILE)
            timeout = True

            while time.time() - start_time < FINI_TIMEOUT:
                if os.path.isfile(finish_path):
                    os.remove(finish_path)
                    timeout = False
                    break
                time.sleep(0.1)

            if timeout:
                raise PE_Close_Timeout()

        # Now, it is true that all parallel writes are fininished.

        attrs = {"vars": {}}

        # restructure data folders
        for proc_path in procs:
            _move_proc(proc_path, self.work_path, attrs)
            shutil.rmtree(proc_path)

        start = {}
        shape = {}

        # merge shape from each proc
        for var_name, var_info in attrs["vars"].items():
            start[var_name], shape[var_name] = _scan(0, var_info["start_length"])
            shape[var_name] = [shape[var_name][-1]] + shape[var_name][:-1]

            if var_info["config"][0]["check"]["shape"] is None:
                continue

            for idx, sh in enumerate(var_info["config"][0]["check"]["shape"]):

                if hasattr(sh, "name"):
                    dim_cfg = self.config["dims"][sh.name]

                    if dim_cfg["length"] is None or dim_cfg["length"] == UNLIMITED:
                        dim_cfg["length"] = shape[var_name][idx]

                    elif dim_cfg["length"] != shape[var_name][idx]:
                        raise PE_Close_Stackdimmismatch("%d != %d" %
                            (dim_cfg["length"], shape[var_name][idx]))

        for name, var_cfg in self.config["vars"].items():

            var_cfg["shape"] = shape[name]
            var_cfg.pop("writes")

            if var_cfg["check"]:
                # check exists if shape arg is given
                for check in var_cfg["check"].keys(): 

                    dim_checks = var_cfg["check"][check]
                    if dim_checks is None:
                        continue

                    if check == "shape":
                        if isinstance(dim_checks, StackDimension):
                            dim_checks.check(shape[name][0])
                            var_cfg["shape"] = test.name

                        elif len(dim_checks) > 0:

                            if isinstance(dim_checks[0], Dimension):
                                dim_checks[0].check(shape[name][0])
                                var_cfg["shape"][0] = dim_checks[0].name
                            elif (dim_checks[0] is not None and dim_checks[0] !=
                                UNLIMITED and shape[name][0] != dim_checks[0]):
                                raise PE_Close_Shapemismatch("%d != %d" %
                                        (shape[name][0], dim_checks[0]))

                            for i, (dim_check, length) in enumerate(
                                    zip(dim_checks[1:], shape[name][1:])):

                                if isinstance(dim_check, Dimension):
                                    dim_check.check(length)
                                    var_cfg["shape"][i+1] = dim_check.name
                                elif (dim_check is not None and dim_check !=
                                    UNLIMITED and shape[name][i+1] != dim_check):
                                    raise PE_Close_Shapemismatch("%d != %d" %
                                            (shape[name][i+1], dim_check))

                    else:
                        raise PE_Close_Unknowncheck(check)

                var_cfg.pop("check")

        slab_path = self.config["_control_"]["slab_path"]

        self.config.pop("_control_")

        pickle_dump(self.cfg_path, self.config)

        with tarfile.open(slab_path, "w") as tar:
            for item in os.listdir(self.work_path):
                item_path = os.path.join(self.work_path, item)
                tar.add(item_path, arcname=item)

        shutil.rmtree(self.work_path)


    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        return self.close()


# non-master implementation of pyslabs 
class ParallelPyslabsWriterV1(PyslabsWriterV1):

    def get_writer(self, name, *vargs, **kwargs):

        var_cfg = self.config["vars"][name]

        if "autostack" in kwargs:
            var_cfg["stack"]["auto"] = kwargs["autostack"]

        return VariableWriterV1(os.path.join(self.proc_path, name), var_cfg)

    def get_dim(self, name):

        dim_cfg = self.config["dims"][name]

        return Dimension(dim_cfg)

    def get_stack(self, name):

        dim_cfg = self.config["dims"][name]

        return StackDimension(dim_cfg)

    def begin(self):
        pass


class PyslabsReaderV1():

    def __init__(self, slab_path):
        self.slab_path = slab_path
        self.tar_file = tarfile.open(slab_path, mode="r:")
        self.slab_tower = OrderedDict()

        tower = {}

        for entry in self.tar_file:
            if entry.name == CONFIG_FILE:
                self.config = pickle.load(self.tar_file.extractfile(entry))

            else:
                self._trie(tower, entry.path.split("/"), entry)

        self._sort_tower(self.slab_tower, tower)


    def _sort_tower(self, dst, src):

        for key in sorted(src.keys()):
            value = src[key]

            if isinstance(value, dict):
                _dst = OrderedDict()
                dst[key] = _dst
                self._sort_tower(_dst, value)

            else:
                dst[key] = value

    def _trie(self, output, entry_path, entry):

        if len(entry_path) == 1:

            if (isinstance(output, tarfile.TarInfo) or
                entry_path[0] in output):
                raise PE_Read_Constructtrie(str(entry_path))

            output[entry_path[0]] = entry

        elif entry_path[0] in output:

            if isinstance(output[entry_path[0]], tarfile.TarInfo):
                _output = {}
                output[entry_path[0]] = _output
                self._trie(_output, entry_path[1:], entry)

            else:
                self._trie(output[entry_path[0]], entry_path[1:], entry)
        else:
            _output = {}
            output[entry_path[0]] = _output
            self._trie(_output, entry_path[1:], entry)

    def get_reader(self, name):

        varcfg = self.config["vars"][name]
        dimcfg = self.config["dims"]

        return VariableReaderV1(self.tar_file, self.slab_tower[name],
                varcfg, dimcfg)

    def get_array(self, name, stack=None):

        if stack is None:
            return self.get_reader(name)[:]

        elif isinstance(stack, (int, slice)):
            return self.get_reader(name).__getitem__(stack)

        else:
            return self.get_reader(name).__getitem__(*stack)


    def close(self):

        if not self.tar_file.closed:
            self.tar_file.close()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def _traverse(self, tree, bag):
            try:
                for k, v in tree.items():
                    if bag["check"](k, v):
                        bag["data"].append(bag["output"](k, v))
                    self._traverse(v, bag)

            except AttributeError as err:
                pass

            except Exception as err:
                import pdb; pdb.set_trace()
                print(err)



    def info(self, mode, *args, **kwargs):

        if mode == "list":
            return tuple(self.slab_tower.keys())

        elif mode == "var":
            out = {}
            var = self.get_reader(args[0])

            import pdb; pdb.set_trace()

        elif mode == "slab":
            out = {}
            bag = {"check": lambda k, v: isinstance(v, tarfile.TarInfo),
                    "output": lambda k, v: v}

            for var, tree in self.slab_tower.items():

                data = []
                bag["data"] = data
                self._traverse(tree, bag)

                if len(data) > 0:
                    totalsize = 0
                    maxsize = 0
                    minsize = sys.maxsize
                    nslabs = len(data)

                    for tar in data:
                        totalsize += tar.size
                        if tar.size > maxsize:
                            maxsize = tar.size
                        if tar.size < minsize:
                            minsize = tar.size
                else:
                    totalsize = maxsize = minsize = nslabs = 0

                out[var] = (nslabs, totalsize, maxsize, minsize)

            return out

        elif mode == "":

            out = []

            out.append(("version", self.config["version"]))

            dbuf = []
            for n, d in self.config["dims"].items():
                dbuf.append((n,  d["length"]))

            out.append(("dims", tuple(dbuf)))

            vbuf = []
            for n, v in self.config["vars"].items():
                if "shape" in v:
                    vbuf.append((n, v["shape"]))
                else:
                    vbuf.append((n, None))

            out.append(("vars", tuple(vbuf)))
            out.append(("size", os.path.getsize(self.slab_path)))

            return out


class ParallelPyslabsReaderV1(PyslabsReaderV1):
    pass


class MasterPyslabsReaderV1(PyslabsReaderV1):
    pass


##############################
# PUBLIC FUNCTIONS
##############################


# open slab I/O for master process
def master_open(slab_path, num_procs, mode="w", workdir=None):

    if mode == "w":

        slab_path, begin_path, work_path = _write_paths(slab_path, workdir)

        # create root directory
        os.makedirs(work_path, exist_ok=True)
        clean_folder(work_path)

        begin = copy.deepcopy(INIT_BEGIN)
        begin["work_path"] = work_path
        begin["slab_path"] = slab_path
        begin["mode"] = "w"

        pickle_dump(begin_path, begin)
        
        # create root directory
#        os.makedirs(work_path, exist_ok=True)
#        clean_folder(work_path)

        config = copy.deepcopy(INIT_CONFIG)
        config["_control_"]["num_procs"] = num_procs
        config["_control_"]["begin_path"] = begin_path
        config["_control_"]["slab_path"] = slab_path

        return MasterPyslabsWriterV1(work_path, config)

    elif mode == "r":

        _cache.clear()

        if num_procs is not None and num_procs > 1:
            print("ERROR: parallel-read is not supported, but 'num_procs' "
                  "argument is larger than one: %d" % num_procs)
            sys.exit(-1)

        return MasterPyslabsReaderV1(slab_path)

    else:
        raise PE_Open_Unknownmode(mode)


# open slab I/O for non-master processes
def parallel_open(slab_path, mode="w"):

    if mode == "w":

        _, begin_path, _ = _write_paths(slab_path, None)

        start = time.time()
        begin = None
        work_path = None

        while time.time() - start < INIT_TIMEOUT:
            if os.path.isfile(begin_path):
                with io.open(begin_path, "rb") as fp:
                    begin = pickle.load(fp)
                    work_path = begin["work_path"]
                break
            time.sleep(0.1)

        if begin is None:
            raise PE_Init_Nobeginfile(begin_path)

        while time.time() - start < INIT_TIMEOUT:
            cfg_path = os.path.join(work_path, CONFIG_FILE)

            if not os.path.isfile(cfg_path):
                time.sleep(0.1)
                continue

            with io.open(cfg_path, "rb") as fp:
                config = pickle.load(fp)

            break

        return ParallelPyslabsWriterV1(work_path, config)

    elif mode == "r":
        return ParallelPyslabsReaderV1(slab_path)

    else:
        raise PE_Open_Unknownmode(mode)


# the wrapper of "master_open" for convinience
def open(slab_path, mode="r", num_procs=1, workdir=None):

    return master_open(slab_path, num_procs, mode=mode, workdir=workdir)



"""Pyslabs constant module


"""

name            = "pyslabs"
version         = "0.1.6"
description     = "Pure Python Paralle I/O"
long_description = """Pure Python Parallel I/O that supports large datasets of
 all pickle-able objects and well-known data frameworks"""
author          = "Youngsung Kim"

UNLIMITED           = -1      # unlimited length on stack dimension
INIT_TIMEOUT        = 10      # seconds
FINI_TIMEOUT        = 100     # seconds
SLAB_EXT            = ".slab" # slab file extension
ZLAB_EXT            = ".zlab" # compressed slab file extension

CONFIG_FILE         = "_config_"
FINISH_FILE         = "_finished_"
VARCFG_FILE         = "_varcfg_"

TMP_BEGIN           = "._tmpbegin_" # an extension of a temporary file
TMP_WORK            = "._tmpwork_" # an extension of a temporary directory

INIT_BEGIN = {
    "work_path": None,
    "slab_path": None,
    "mode": None 
}

INIT_CONFIG = {
    "version": 1,
    "dims": {},
    "vars": {},
    "attrs": {},
    "_control_": {
    }
}

INIT_VARCFG = {
    "writes": {},
    "shape": None,
    "check": {},
    "attrs": {},
    "stack": {}
}

INIT_DIMCFG = {
    "attrs": {}
}





"""pyslabs info module
"""

name            = "pyslabs"
version         = "0.1.3"
description     = "Pure Python Paralle I/O"
long_description = """Pure Python Parallel I/O that supports large datasets of
 all pickle-able objects and well-known data frameworks"""
author          = "Youngsung Kim"



"""Pyslabs slab write module


"""

import os
from pyslabs import slabif
from pyslabs.util import arraytype
from pyslabs.error import PE_Slab_Shapemismatch, PE_Write_Duplicateslabfile


class VariableWriterV1():

    def __init__(self, path, config):

        self.path = path
        self.config = config
        self.check_shape = config["check"]["shape"]
        self.auto_stack = config["stack"]["auto"]
        self.level = 0

    def stacking(self, nlevel=1):
        self.level += nlevel

    def write_panel(self, panel, start=None, level=None):

        # TODO: check if panel is multiples of each slab dims
        # TODO: split panel into multiple slabs
        # TODO: generate start and level per each slabs
        # TODO: writes slabs
        pass

    def write(self, slab, start=None, level=None):

        # get slab info
        slab_shape = slabif.shape(slab)

        # normalize start 
        if start is None:
            start = (0,) * len(slab_shape)

        elif isinstance(start, int):
            start = (start,) + (0,) * (len(slab_shape) - 1)

        else:
            start = start + (0,) * (len(slab_shape) - len(start))

        # generate relative path to data file
        rel_path = []

        for st, sh in zip(start, slab_shape):
            rel_path.append(str(st)+"_"+str(sh))

        strlevel = str(self.level) if level is None else str(level)

        if strlevel in self.config["writes"]:
            writes = self.config["writes"][strlevel]

        else:
            writes = {}
            self.config["writes"][strlevel] = writes

        writes["/".join(rel_path)] = (start, slab_shape)

        slab_folder = os.path.join(self.path, *rel_path)

        if not os.path.isdir(slab_folder):
            os.makedirs(slab_folder)

        atype, ext = arraytype(slab)
        slab_path = os.path.join(slab_folder, ".".join([strlevel, atype, ext]))

        if os.path.isfile(slab_path):
            raise PE_Write_Duplicateslabfile(slab_path)

        slabif.dump(slab_path, slab)

        if level is None:
            if self.auto_stack is True:
                self.stacking()

            elif self.auto_stack > 0:
                self.stacking(nlevel=self.auto_stack)
